#
#   Entry Points:  New, load_config, process
#
#   This module takes a raw file, and converts it to a PIF file. Functionality
#   includes:
#   
# (c) Copyright IBM Corporation 2007. All rights reserved.
#
package SIEMENS_INTERFACE;

use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);

use AudLog;
use GenUtils;
use PIF_Handler;
use File::Basename;

require Exporter;
require AutoLoader;

@ISA = qw(Exporter AutoLoader);
# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.
@EXPORT = qw();
$VERSION = '0.01';


my $config_cmd = {
	LINE => {
		PATTERN	=> '^(LINE)\s+(\w+)',
		HANDLER	=> \&invoke_cmd_line
		},
	DATAB => {
		PATTERN	=> '^(DATAB)\s+(\w+)\s+(\d+)\s+(\w+)',
		HANDLER	=> \&invoke_cmd_datab
		},
	DATABNP	=> {
		PATTERN => '^(DATABNP)\s+(\w+)\s+(\d+)\s+(\w+)' ,
		HANDLER	=> \&invoke_cmd_datab
		},						
	DATAO => {
		PATTERN => '^(DATAO)\s+(\w+)\s+(\d+)\s+(\d+)\s+(\w+)',
		HANDLER	=> \&invoke_cmd_datao
		},
	DATAONP	=> {
		PATTERN	=> '^(DATAONP)\s+(\w+)\s+(\d+)\s+(\d+)\s+(\w+)',
		HANDLER	=> \&invoke_cmd_datao
		},							
	REPEATEND => {
		PATTERN	=> '^(REPEATEND) \{\s*((.|\n)*?) \} REPEATEND',
		HANDLER	=> \&invoke_cmd_repeatend
		},
	JOIN => {	
		PATTERN	=> '^(JOIN)\s+(\w+)\s+(.*\w+.*)',
		HANDLER	=> \&invoke_cmd_join
		},	
	CONTROL => {
		PATTERN => '^(CONTROL)\s+(\w+)',
		HANDLER	=>  \&invoke_cmd_control
		},
};

my (%config_data, %counter_names, %join_buf, %ctrl_buf);

my $data_buf;
my $data_len;
my $data_pos;

my $line_pos;


################################################################################
# Subroutine name:  New()
#
# Description:      Object initialisation routine
#
# Arguments:        None
#
# Returns:          Object reference
#
sub New {
	my $proto = shift;
	my $class = ref($proto) || $proto;
	my $self = {};

	# These variables should be set to true of false
	# depending on whether or not the desired type of output file
	# is wanted
	$self->{PRODUCE_PIF} = 'True';

	bless ($self, $class);
}

################################################################################
# Subroutine name:  load_config()
#
# Description:      Object configuration loading routine
#
# Arguments:        keep_input_dir(scalar) - indicating where to store input file 
#                                         once it has been processed (if defined). 
#                   debug (scalar) - a boolean indicating whether or not the
#                                    parser is being run in debug mode.
#                   config (scalar) - a reference to a hash that contains all
#                                     the configuration options that have to
#                                     be loaded.
#
# Returns:          0 for success,
#                   the number of errors found for failure
#
sub load_config {
	my $self = shift;
	($self->{keep_input_dir}, $self->{debug}, $self->{'__config__'}) = @_;

	# Inserting all the configuration information into the objects records
	my ($key, $num_errors);
	foreach $key ( keys %{$self->{'__config__'}} ) {
		$self->{$key} = $self->{'__config__'}->{$key};
	}

	$num_errors = 0;

	# CHECK MANDATORY OPTIONS are present
	# validate the configuration.
	my %mandatory = (
		INPUT_FILE_DESCRIPTION		=> ['SCALAR', 'ARRAY'],
		CONFIG_FILENAME               	=> 'SCALAR',
         );
    
	my %optional = (
		HEADER_FIELDS_FOR_PIF_FILENAME    	=> 'ARRAY',
		DATA_FIELDS_FOR_PIF_FILENAME      	=> 'ARRAY',
		HEADER_DATA_RECORD_PROCESSING     	=> 'CODE',
         );
  
	$num_errors = GenUtils::check_configuration($self->{'__config__'},
                                                \%mandatory,\%optional);
						
	if ($num_errors){
		LogMess("ERROR checking configuration $num_errors errors in config",3);
		return -1;
	}
    
	if ($self->load_parser_config()) {
		LogMess("ERROR reading configuration file $self->{CONFIG_FILENAME}",3);
		return -1;    
	}

	return $num_errors;
}


################################################################################
# Subroutine name:  process_file()
#
# Description:      Controls the processing of all the files that match the
#                   INPUT_FILE_DESCRIPTION field in the config information
#
# Arguments:        self (scalar) - a reference to a hash that contains all 
#                                   the configuration options for this process.
#                                   
#                   filenm (scalar) - the location and filename of the file to be 
#                   processed. 
#
#	            headers(scalar) - a reference to a hash that contains filename
#		    and directory name extractions.
#
# Returns:          (scalar) - 0 = success
#	       	    -1 = Failed, rename input file to '.bad' 
#		    -2 = Failed, do not rename input file to '.bad'
#
sub process_file {
	my ($self, $filenm, $headers) =@_;
	my ( %header_fields, %data_fields, $validity_field, $invalid_field );
	my ( @filename_fields, @header_counters, @data_counters, $invalid_value );
    
	my ($valid_no, $valid_class, $turntobad) = -1;
	
	# Establish the directory that outputted files will be stored in.
	my $int_dir = $self->{OUTPUT_DIR};

	# Read in the raw file
	$self->{FILE} = open_file($filenm);
    
	if ( $self->{FILE} < 0) {
		LogMess("SIEMENS_NSS: ERROR could not open raw file $filenm", 2);
		return -1;
	}
	
	# Extract filename from the file reference
	my $fname = basename($filenm);
	LogMess("SIEMENS_NSS: Processing PM file $fname", 5);
    
	my $raw_fh = $self->{FILE};
	
	# Extract the first line of the raw file and parse for header information
	%counter_names = ();
	%join_buf = ();
	%ctrl_buf = ();
	
	$line_pos = 0;
	
	process_header_block( \%header_fields, \@header_counters, \%data_fields, \@data_counters, $raw_fh);

 	my %common_h = ();

	foreach (keys %$headers) {
		$common_h{$_} = $headers->{$_}; 
	}
	
 	my $pif_header = shift @header_counters;

	foreach (keys %$pif_header) {
		$common_h{$_} = $pif_header->{$_};		
	} 
	
	close($raw_fh);
	
	$self->write_pif_cache(\%common_h, \@data_counters);
	
	return 0;
}


sub write_pif_cache {
	my ( $self, $header_ref, $data_ref ) = @_;
	my ( @filename_fields, $header_fields, $data_fields, $pif_obj, $pif_ctx );
	my ( $filename, $startdate, $starttime);
	
	LogMess("Clearing out PIF data cache", 5);
	
	($pif_obj, $pif_ctx) = GenUtils::get_pif_object();
	
	# Generating PIF filename
	
	$header_fields = $header_ref;
	$data_fields = $data_ref->[0];
	
	$startdate = $data_fields->{START_DATE};
	$starttime = $data_fields->{TIME};

	foreach my $field (@{$self->{HEADER_FIELDS_FOR_PIF_FILENAME}}) {
		if (exists($header_fields->{$field})) {
			push @filename_fields, $header_fields->{$field};
		}
	}
	
	unless ($filename) {
		$filename = join("-#-", @filename_fields, $startdate , $starttime,"I.pif");
	}
	
	$pif_obj->Open_PIF_Write($pif_ctx,$self->{OUTPUT_DIR} . '/' . $filename);
	
	$pif_obj->PIF_Write($pif_ctx,'HEADER', $header_fields);
	
	foreach my $pif_row (@$data_ref) {
		foreach (keys %$pif_row) {
			if (exists($counter_names{$_})) {
				$counter_names{$_} = $pif_row->{$_};
			}
		}
	
		$pif_obj->PIF_Write($pif_ctx, $header_fields->{MEASURE_ID}, \%counter_names);

		foreach (keys %counter_names) {
			$counter_names{$_} = '';
		}
	}
	
	$pif_obj->Close_PIF_Write($pif_ctx);

	LogMess("Written $filename to pif", 5);
}


# Handler for DATAB command which extracts data. 
sub invoke_cmd_datab {
	my ($cmd_ref, $counters_ref, $cache_ref, $fh) = @_;
	my ($name_ref, $length_ref, $type_ref, $value);
	
	if ($data_pos > length($data_buf)) {
		return 0;
	}

	LogMess("Invoking DATAB command", 5);
		
	$name_ref = \$cmd_ref->{COUNTER};
	$length_ref = \$cmd_ref->{LENGTH};
	$type_ref = \$cmd_ref->{TYPE};
	
	$value = substr($data_buf, $data_pos, $$length_ref);
	
	if ($cmd_ref->{COMMAND} !~ /^DATABNP$/) {
		$value = trim($value);
		check_data_type(\$value, $type_ref);
		$counters_ref->{$$name_ref} = $value;
		$counter_names{$$name_ref} = '';
	}
	
	$data_pos += $$length_ref;

	LogMess("DEBUG: Line Position=" . $data_pos , 5);
	LogMess("DEBUG: " . $$name_ref . "=" . $counters_ref->{$$name_ref}, 5);
	
	return 0;
}


# Handler for DATAO command which extracts data.
sub invoke_cmd_datao {
	my ($cmd_ref, $counters_ref, $cache_ref, $fh) = @_;
	my ($name_ref, $length_ref, $offset_ref, $type_ref, $value); 
	
	LogMess("Invoking DATAO command", 5);
	
	$name_ref = \$cmd_ref->{COUNTER};
	$offset_ref = \$cmd_ref->{OFFSET};
	$length_ref = \$cmd_ref->{LENGTH};
	$type_ref = \$cmd_ref->{TYPE};


	$value = substr($data_buf, $$offset_ref - 1, $$length_ref);

	if ($cmd_ref->{COMMAND} !~ /^DATAONP$/) {
		$value = trim($value);
		check_data_type(\$value, $type_ref);
		$counters_ref->{$$name_ref} = $value;
		$counter_names{$$name_ref} = '';
	}
	
	LogMess("DEBUG: " . $$name_ref . "=" . $counters_ref->{$$name_ref}, 5);
	
	return 0;
}


sub invoke_cmd_repeatend {
	my ($cmd_ref, $counters_ref, $cache_ref, $fh) = @_;
	
	LogMess("Invoking REPEATEND command", 5);
	LogMess("DEBUG: DATAPOS -> " . $data_pos . "/" . length($data_buf), 5);

	my $cmd_data = $cmd_ref->{DATA};
	
	my %row_header = %$counters_ref;
	
	foreach my $cmd_line (@$cmd_data) {
		$config_cmd->{$cmd_line->{COMMAND}}->{HANDLER}->($cmd_line, $counters_ref, $cache_ref);
	}
	
	if ( $data_pos > length($data_buf)) {
		return 0;
	}
	
	my %tmp_hash = %$counters_ref;
	
	if (keys(%tmp_hash)) {
		push @$cache_ref, \%tmp_hash;
	}
	
	%$counters_ref = ();
	
	if ($data_pos < (length($data_buf)-1)) {
		 invoke_cmd_repeatend($cmd_ref, \%row_header, $cache_ref);
	} else {
		return 0;
	}

}


sub invoke_cmd_join {
	my ($cmd_ref, $counters_ref, $cache_ref, $fh) = @_;
	
	LogMess("Invoking JOIN command", 5);

	if ($cmd_ref->{COUNTER} =~ /^"(.*)"$/) {
		$join_buf{$cmd_ref->{BUFFER}} .= $1;    
	} else {
		$join_buf{$cmd_ref->{BUFFER}} .= $counters_ref->{$cmd_ref->{COUNTER}};		
	}
	
	LogMess("DEBUG: " . $cmd_ref->{BUFFER} . " = " . $join_buf{$cmd_ref->{BUFFER}}, 5);
	
	return 0;
}


sub invoke_cmd_control {
	my ($cmd_ref, $counters_ref, $cache_ref, $fh) = @_;
	my ($block_name, $buffer_name);

	LogMess("Invoking CONTROL command", 5);
	
	my @buffer_list = split(/\s+/, $cmd_ref->{BUFFER});
	$buffer_name =  shift @buffer_list;
	
	if (exists($join_buf{$buffer_name}) ) {
		$ctrl_buf{BUFFER} = $join_buf{$buffer_name};
		delete $join_buf{$buffer_name};
	}
	
	my $line_data = $config_data{$ctrl_buf{BUFFER}};
	foreach my $cmd_data (@$line_data) {
		foreach my $cmd_line (@$cmd_data) {
			if ($cmd_line->{COMMAND} =~ /^LINE$/) {
				$config_cmd->{'LINE'}->{HANDLER}->($fh);	
			}
			else {
				$config_cmd-> {$cmd_line->{COMMAND}}->{HANDLER}->($cmd_line, $counters_ref, $cache_ref);
			}
		}
	}
	
	LogMess("DEBUG: CONTROL BUFFER = " . $ctrl_buf{BUFFER}, 5);
	
	return 0; 
}


sub invoke_cmd_line {
	my ($fh) = @_;
	
	$line_pos++;
	
	LogMess("Invoking LINE command, LINE " . $line_pos, 5);
	
	$data_buf = <$fh>;

	if ($data_buf =~ /^VS/) {
		$data_buf = <$fh>;
	}
	
	$data_pos = 0;
	
	return 0;
}


sub process_header_block {
	my ($hfld_ref, $hrow_ref, $dfld_ref, $drow_ref, $fh) = @_;
	
	LogMess("Processing HEADER Block", 3);
	
	my $line_data = $config_data{'HEADER'};
	
	foreach my $cmd_data (@$line_data) {
		foreach my $cmd_line (@$cmd_data) {
			if ($cmd_line->{COMMAND} =~ /^LINE$/) {
				$config_cmd->{'LINE'}->{HANDLER}->($fh);	
			}
			elsif ($cmd_line->{COMMAND} =~ /^CONTROL$/) {
				my $buffer_name =  $cmd_line->{BUFFER};
					
				if (exists($join_buf{$buffer_name}) ) {
					process_config_block($join_buf{$buffer_name}, $hfld_ref, $hrow_ref, $dfld_ref, $drow_ref, $fh);
					delete $join_buf{$buffer_name};
				}
			}
			else {
				$config_cmd-> {$cmd_line->{COMMAND}}->{HANDLER}->($cmd_line, $hfld_ref, $hrow_ref, $fh);
			}
		}
		
		push @$hrow_ref, $hfld_ref;
	}

	return 0;
}


sub process_config_block {
	my ($config_name, $hfld_ref, $hrow_ref, $dfld_ref, $drow_ref, $fh) = @_;
	
	LogMess("Processing Config Block " . $config_name, 3);
	
	if (!exists($config_data{$config_name})) { 	
		LogMess("Unable to find Config Block $config_name", 3);
		return 1;
	}
	
	%counter_names = ();
	
	my $line_data = $config_data{$config_name};

	LINE : while (!eof($fh)) {
		foreach my $cmd_data (@$line_data) {
			foreach my $cmd_line (@$cmd_data) {
				if ($cmd_line->{COMMAND} =~ /^LINE$/) {
					$config_cmd->{'LINE'}->{HANDLER}->($fh);	

				(next LINE) if (trim($data_buf) <= 0);
				}
				else {
					$config_cmd-> {$cmd_line->{COMMAND}}->{HANDLER}->($cmd_line, $dfld_ref, $drow_ref, $fh);
				}
			}
			
			if (keys(%$dfld_ref)) {
				my %temp_hash = %$dfld_ref;
				push @$drow_ref, \%temp_hash
				# push @$drow_ref, $dfld_ref;
			}
		}
	}

	return 0;
}


sub populate_cmd_data {
	
	my $cmd_str = @_[0];
	
	my @cmd_data = ();
	my @tmp_data = ();
	my $tmp_str = '';
	
	while ($cmd_str =~ /((DATAB|DATABNP|DATAO|DATAONP|CONTROL|JOIN|LINE).*\n|REPEATEND \{\s*((.|\n)*?) \} REPEATEND)/g) {
		$tmp_str = trim($1);
		push(@tmp_data, $tmp_str);
	}

	foreach my $data (@tmp_data) {
		
		SWITCH: {
			$data =~ /^$config_cmd->{LINE}->{PATTERN}/ and do {
				$data =~ s/^$config_cmd->{LINE}->{PATTERN}/COMMAND=$1;FLAG=$2/g;
				push @cmd_data, {split(/[=;]/, $data)};
				last;
			};
			$data =~ /^$config_cmd->{DATAB}->{PATTERN}/ and do {
				$data =~ s/^$config_cmd->{DATAB}->{PATTERN}/COMMAND=$1;COUNTER=$2;LENGTH=$3;TYPE=$4/g;
				push @cmd_data, {split(/[=;]/, $data)};
				last;
			};
			$data =~ /^$config_cmd->{DATABNP}->{PATTERN}/ and do {
				$data =~ s/^$config_cmd->{DATABNP}->{PATTERN}/COMMAND=$1;COUNTER=$2;LENGTH=$3;TYPE=$4/g;
				push @cmd_data, {split(/[=;]/, $data)};
				last;
			};
			$data =~ /^$config_cmd->{DATAO}->{PATTERN}/ and do {
				$data =~ s/^$config_cmd->{DATAO}->{PATTERN}/COMMAND=$1;COUNTER=$2;OFFSET=$3;LENGTH=$4;TYPE=$5/g;
				push @cmd_data, {split(/[=;]/, $data)};
				last;
			};
			$data =~ /^$config_cmd->{DATAONP}->{PATTERN}/ and do {
				$data =~ s/^$config_cmd->{DATAONP}->{PATTERN}/COMMAND=$1;COUNTER=$2;OFFSET=$3;LENGTH=$4;TYPE=$5/g;
				push @cmd_data, {split(/[=;]/, $data)};
				last;
			};
			$data =~ /^$config_cmd->{REPEATEND}->{PATTERN}/ and do {
				push @cmd_data, {COMMAND => 'REPEATEND', DATA =>  populate_cmd_data($2)};
				last;
			};
			$data =~ /$config_cmd->{JOIN}->{PATTERN}/ and do {
				$data =~ s/^$config_cmd->{JOIN}->{PATTERN}/COMMAND=$1;BUFFER=$2;COUNTER=$3/;
				push @cmd_data, {split(/[=;]/, $data)};
				last;
			};
			$data =~ /$config_cmd->{CONTROL}->{PATTERN}/ and  do {
				$data =~ s/^$config_cmd->{CONTROL}->{PATTERN}/COMMAND=$1;BUFFER=$2/;
				push @cmd_data, {split(/[=;]/, $data)};
				last;
			};
		}
	}
	
	return \@cmd_data;
}


sub populate_line_data {
	
	my ($block_name, $config_block, $line_data) = @_;

	$config_block = "LINE True\n" . $config_block;
	
	push @$line_data, populate_cmd_data($config_block);

	return 0;
}


sub init_config_block {
	
	my ($block_name, $fh) = @_;
	 
	my $config_block = '';
	my @line_data;
	
	while ((my $line = <$fh>) !~ /\} ($block_name)/) {
		$config_block .=  $line;
	}
	
	LogMess("DEBUG: Parsing Config Block $block_name", 5);
	LogMess("DEBUG: Parsing Config Block data -> \n" . $config_block, 5);
		
	while ($config_block  =~ /LINE\s*\{(((?! LINE\s*{).)*?)\}\s*LINE/gsi) {
		populate_line_data  $block_name, $1, \@line_data;
		$config_data{$block_name} = \@line_data;
	}
	
	if ($config_block !~  /LINE\s*\{(((?! LINE\s*{).)*?)\}\s*LINE/gsi) {
		push @line_data, populate_cmd_data($config_block);
		$config_data{$block_name} = \@line_data;
	}
	
	return 0;
}


sub parse_conf_file {
	
	my $conf_fh = @_[0];	
	
	%config_data = ();
	
	while (<$conf_fh>) {
		if ( $_ =~ /^\s*\{ (\w+)/ ) {		
			init_config_block($1, $conf_fh);
		}
	}
    
	close($conf_fh);
    
	return 0;
}


sub load_parser_config {
 
	my $self = shift;
    
	LogMess("Loading parser configuration file from " . $self->{CONFIG_FILENAME}, 5);

	my $config_fh = &GenUtils::open_file($self->{CONFIG_FILENAME});

	if ( $config_fh < 0 ) {
		LogMess("ERROR: Failed to open config file $self->{CONFIG_FILENAME}",2);
		return -2;
	}
	
	parse_conf_file($config_fh);
 
	LogMess("Parser configuration successfully loaded", 5);
 
	return 0;
}


sub process_data_quality {
	
	my $value_ref = shift;
	
	if ($$value_ref != 1) {
		LogMess("BAD DATA QUALITY on line $line_pos", 3);
	}
	
	return 0;
}


sub process_secure_mark {
	
	my $value_ref = shift;
	
	if (length($$value_ref) != 0) {
		LogMess("BAD SECURE MARK on line $line_pos", 3);
	}
	
	return 0;
}


sub check_data_type {
	
	my ($value_ref, $type_ref) = @_;
	
	SWITCH: {
		$$type_ref =~ /INT/ and do {
			last;
		};
		$$type_ref =~ /REAL/ and do {
			last;
		};
		$$type_ref =~ /STRING/ and do {
			last;
		};
		$$type_ref =~ /NONE/ and do {
			last;
		};
		$$type_ref =~ /DATE_FORMAT/ and do {
			GenUtils::convert_date_format($value_ref);
			last;
		};
		$$type_ref =~ /DATA_LENGTH/ and do {
			last;
		};
		$$type_ref =~ /DATA_QUALITY/ and do {
			process_data_quality($value_ref);
			last;
		};
		$$type_ref =~ /SECURE_MARK/ and do {
			process_secure_mark($value_ref);
			last;
		};	
	};
	
	return 0;
}


sub trim {
	
	my $string_ref = shift @_;
	
	$string_ref =~ s/^\s+//;   #Trim spaces at the start of the string
	$string_ref =~ s/\s+$//;   # Trim spaces at the end of the string
	
	return $string_ref;	
}

1;  # tells perl that the package is ready to run
